@@ -9,8 +9,18 @@
 
 from dataclasses import dataclass, field
 
-from rdflib import (DC, FOAF, PROV, RDF, RDFS, XSD, Graph,  # type: ignore
-                    Literal, Namespace, URIRef)
+from rdflib import (
+    DC,
+    FOAF,
+    PROV,
+    RDF,
+    RDFS,
+    XSD,
+    Graph,  # type: ignore
+    Literal,
+    Namespace,
+    URIRef,
+)
 from validators import url
 
 from provo.idvault import IdVault
@@ -47,7 +57,7 @@ class PrefixNotAllowed(Exception):
 
 
 @dataclass
-class ProvOntologyGraph():
+class ProvOntologyGraph:
     """model that manages contents of a provenance graph
     that adheres to the PROV-O provenance model."""
 
@@ -65,36 +75,77 @@ class ProvOntologyGraph():
         # validate namespace
         # TODO rethink validation
         if not url(self.namespace):  # type: ignore
-            raise NamespaceMalformed("""
+            raise NamespaceMalformed(
+                """
             The provided namespace is not a valid URL!
 
             See validators package (https://github.com/python-validators/validators),
             validators.url() or https://gist.github.com/dperini/729294 for more
-            information on what qualifies a URL as valid.""")
+            information on what qualifies a URL as valid.
+            
+            https://www.rfc-editor.org/rfc/rfc3986#section-3:
+            3.  Syntax Components
+
+            The generic URI syntax consists of a hierarchical sequence of
+            components referred to as the scheme, authority, path, query, and
+            fragment.
+
+                URI         = scheme ":" hier-part [ "?" query ] [ "#" fragment ]
+
+                hier-part   = "//" authority path-abempty
+                            / path-absolute
+                            / path-rootless
+                            / path-empty
+
+            The scheme and path components are required, though the path may be
+            empty (no characters).  When authority is present, the path must
+            either be empty or begin with a slash ("/") character.  When
+            authority is not present, the path cannot begin with two slash
+            characters ("//").  These restrictions result in five different ABNF
+            rules for a path (Section 3.3), only one of which will match any
+            given URI reference.
+
+            The following are two example URIs and their component parts:
+
+                    foo://example.com:8042/over/there?name=ferret#nose
+                    \_/   \______________/\_________/ \_________/ \__/
+                    |           |            |            |        |
+                scheme     authority       path        query   fragment
+                    |   _____________________|__
+                    / \ /                        \
+                    urn:example:animal:ferret:nose
+            """
+            )
         end_symbol = self.namespace[-1]
-        if end_symbol not in ('/', '#'):
-            raise NamespaceHasNoEndSymbol("The provided Namespace has to end on a '/' or '#'!")
+        if end_symbol not in ("/", "#"):
+            raise NamespaceHasNoEndSymbol(
+                "The provided Namespace has to end on a '/' or '#'!"
+            )
 
         # validate namespace abbreviation
         # TODO check what is actually allowed
-        allowed_symbols_for_prefix_shorthand = 'abcdefghijklmnopqrstuvwxyz'
+        allowed_symbols_for_prefix_shorthand = "abcdefghijklmnopqrstuvwxyz"
 
         for symbol in self.namespace_abbreviation:
             if symbol not in allowed_symbols_for_prefix_shorthand:
-                raise PrefixShorthandNotValid(f"""
+                raise PrefixShorthandNotValid(
+                    f"""
                     The provided namespace abbreviation is not valid.
                     
                     Character of the namespace have to be in "{allowed_symbols_for_prefix_shorthand}".
-                    """)
+                    """
+                )
         # if self.namespace_abbreviation in forbidden_namespaces:
         core_prefixes = ["owl", "rdf", "rdfs", "xsd", "xml"]
         if self.namespace_abbreviation in core_prefixes:
-            raise PrefixNotAllowed(f"""
+            raise PrefixNotAllowed(
+                f"""
             The provided namespace abbreviation is a core prefix, 
             and thus, prohibited from use.
 
             Core prefixes are: {"".join([f"{k}, " if i != len(core_prefixes)-2 else f"{k} and " for i, k in enumerate(core_prefixes)])[:-2]}.
-            """)
+            """
+            )
 
     def __str__(self) -> str:
         """prints the contents of the provenance graph in a nice format."""
@@ -138,36 +189,45 @@ class ProvOntologyGraph():
             node_id = self._id_vault.generate(namespace)
         return node_id
 
-    def add_entity(self, id_string: str = "", label: str = "", description: str = "", namespace: str = "") -> Entity:
+    def add_entity(
+        self,
+        id_string: str = "",
+        label: str = "",
+        description: str = "",
+        namespace: str = "",
+    ) -> Entity:
         """creates a new entity, adds it to the graph and returns it then"""
 
         node_id = self._handle_id(namespace, id_string)
-        entity = Entity(
-            node_id=node_id,
-            label=label,
-            description=description)
+        entity = Entity(node_id=node_id, label=label, description=description)
         self._entities.append(entity)
         return entity
 
-    def add_activity(self, id_string: str = "", label: str = "", description: str = "", namespace: str = "") -> Activity:
+    def add_activity(
+        self,
+        id_string: str = "",
+        label: str = "",
+        description: str = "",
+        namespace: str = "",
+    ) -> Activity:
         """creates a new activity, adds it to the graph and returns it then"""
 
         node_id = self._handle_id(namespace, id_string)
-        activity = Activity(
-            node_id=node_id,
-            label=label,
-            description=description)
+        activity = Activity(node_id=node_id, label=label, description=description)
         self._activities.append(activity)
         return activity
 
-    def add_agent(self, id_string: str = "", label: str = "", description: str = "", namespace: str = "") -> Agent:
+    def add_agent(
+        self,
+        id_string: str = "",
+        label: str = "",
+        description: str = "",
+        namespace: str = "",
+    ) -> Agent:
         """creates a new agent, adds it to the graph and returns it then"""
 
         node_id = self._handle_id(namespace, id_string)
-        agent = Agent(
-            node_id=node_id,
-            label=label,
-            description=description)
+        agent = Agent(node_id=node_id, label=label, description=description)
         self._agents.append(agent)
         return agent
 
@@ -190,61 +250,95 @@ class ProvOntologyGraph():
 
         for node in self._entities + self._activities + self._agents:
             if node.label:
-                provenance_graph.add((
-                    URIRef(node.node_id), RDFS.label, Literal(node.label, lang=self.lang)
-                ))
+                provenance_graph.add(
+                    (
+                        URIRef(node.node_id),
+                        RDFS.label,
+                        Literal(node.label, lang=self.lang),
+                    )
+                )
             if node.description:
-                provenance_graph.add((
-                    URIRef(node.node_id), RDFS.comment, Literal(node.description, lang=self.lang)
-                ))
+                provenance_graph.add(
+                    (
+                        URIRef(node.node_id),
+                        RDFS.comment,
+                        Literal(node.description, lang=self.lang),
+                    )
+                )
         for entity in self._entities:
-            provenance_graph.add((
-                URIRef(entity.node_id), RDF.type, PROV.Entity
-            ))
+            provenance_graph.add((URIRef(entity.node_id), RDF.type, PROV.Entity))
             for activity in entity._was_generated_by_activities:
-                provenance_graph.add((
-                    URIRef(entity.node_id), PROV.wasGeneratedBy, URIRef(activity.node_id)
-                ))
+                provenance_graph.add(
+                    (
+                        URIRef(entity.node_id),
+                        PROV.wasGeneratedBy,
+                        URIRef(activity.node_id),
+                    )
+                )
             for origin_entity in entity._was_derived_from_entities:
-                provenance_graph.add((
-                    URIRef(entity.node_id), PROV.wasDerivedFrom, URIRef(origin_entity.node_id)
-                ))
+                provenance_graph.add(
+                    (
+                        URIRef(entity.node_id),
+                        PROV.wasDerivedFrom,
+                        URIRef(origin_entity.node_id),
+                    )
+                )
             for agent in entity._was_attributed_to_agents:
-                provenance_graph.add((
-                    URIRef(entity.node_id), PROV.wasAttributedTo, URIRef(agent.node_id)
-                ))
+                provenance_graph.add(
+                    (
+                        URIRef(entity.node_id),
+                        PROV.wasAttributedTo,
+                        URIRef(agent.node_id),
+                    )
+                )
         for activity in self._activities:
-            provenance_graph.add((
-                URIRef(activity.node_id), RDF.type, PROV.Activity
-            ))
+            provenance_graph.add((URIRef(activity.node_id), RDF.type, PROV.Activity))
             if activity._start_time:
-                provenance_graph.add((
-                    URIRef(activity.node_id), PROV.startedAtTime, Literal(activity._start_time, datatype=XSD.dateTime)
-                ))
+                provenance_graph.add(
+                    (
+                        URIRef(activity.node_id),
+                        PROV.startedAtTime,
+                        Literal(activity._start_time, datatype=XSD.dateTime),
+                    )
+                )
             if activity._end_time:
-                provenance_graph.add((
-                    URIRef(activity.node_id), PROV.endedAtTime, Literal(activity._end_time, datatype=XSD.dateTime)
-                ))
+                provenance_graph.add(
+                    (
+                        URIRef(activity.node_id),
+                        PROV.endedAtTime,
+                        Literal(activity._end_time, datatype=XSD.dateTime),
+                    )
+                )
             for entity in activity._used_entities:
-                provenance_graph.add((
-                    URIRef(activity.node_id), PROV.used, URIRef(entity.node_id)
-                ))
+                provenance_graph.add(
+                    (URIRef(activity.node_id), PROV.used, URIRef(entity.node_id))
+                )
             for previous_activity in activity._was_informed_by_activities:
-                provenance_graph.add((
-                    URIRef(activity.node_id), PROV.wasInformedBy, URIRef(previous_activity.node_id)
-                ))
+                provenance_graph.add(
+                    (
+                        URIRef(activity.node_id),
+                        PROV.wasInformedBy,
+                        URIRef(previous_activity.node_id),
+                    )
+                )
             for agent in activity._was_associated_with_agents:
-                provenance_graph.add((
-                    URIRef(agent.node_id), PROV.wasAssociatedWith, URIRef(agent.node_id)
-                ))
+                provenance_graph.add(
+                    (
+                        URIRef(agent.node_id),
+                        PROV.wasAssociatedWith,
+                        URIRef(agent.node_id),
+                    )
+                )
         for agent in self._agents:
-            provenance_graph.add((
-                URIRef(agent.node_id), RDF.type, PROV.Agent
-            ))
+            provenance_graph.add((URIRef(agent.node_id), RDF.type, PROV.Agent))
             for instructor in agent._acted_on_behalf_of_agents:
-                provenance_graph.add((
-                    URIRef(agent.node_id), PROV.actedOnBehalfOf, URIRef(instructor.node_id)
-                ))
+                provenance_graph.add(
+                    (
+                        URIRef(agent.node_id),
+                        PROV.actedOnBehalfOf,
+                        URIRef(instructor.node_id),
+                    )
+                )
 
         return provenance_graph
 
@@ -254,19 +348,21 @@ class ProvOntologyGraph():
         """
         self.get_rdflib_graph().serialize(destination=file_name, format=export_format)
 
-    def export_as_mermaid_flowchart(self, file_name: str, user_options: dict = {}) -> None:
+    def export_as_mermaid_flowchart(
+        self, file_name: str, user_options: dict = {}
+    ) -> None:
         """exports the contents of the graph as mermaid-md flowchart"""
 
         # if possible the options use the mermaid terminology
         options = {
             "revert-relations": False,
-            "orientation": "TD",            
+            "orientation": "TD",
             "included-relations": [
                 "was_generated_by",
                 "was_attributed_to",
                 "used",
                 "was_associated_with",
-                "acted_on_behalf_of"
+                "acted_on_behalf_of",
             ],
             "color": "#000000",
             "stroke": "#a4a4a4",
@@ -278,7 +374,7 @@ class ProvOntologyGraph():
                 "relation-style": None,
                 "color": None,
                 "stroke": None,
-                "stroke-width": None
+                "stroke-width": None,
             },
             "activity": {
                 "fill": "#cfceff",
@@ -286,16 +382,16 @@ class ProvOntologyGraph():
                 "relation-style": None,
                 "color": None,
                 "stroke": None,
-                "stroke-width": None
+                "stroke-width": None,
             },
             "agent": {
                 "fill": "#ffebc3",
                 "shape": "[/:\\]",
                 "relation-style": ".",
                 "color": None,
-                "stroke":  None,
-                "stroke-width": None
-            }
+                "stroke": None,
+                "stroke-width": None,
+            },
         }
 
         options = update_dict(options, user_options)
@@ -334,7 +430,9 @@ class ProvOntologyGraph():
         elif stroke_color:
             lines.append(f"classDef entity stroke:{options['stroke']}")
         if entity_stroke_width:
-            lines.append(f"classDef entity stroke-width:{options['entity']['stroke-width']}")
+            lines.append(
+                f"classDef entity stroke-width:{options['entity']['stroke-width']}"
+            )
         elif stroke_width:
             lines.append(f"classDef entity stroke-width:{options['stroke-width']}")
 
@@ -348,7 +446,9 @@ class ProvOntologyGraph():
         elif stroke_color:
             lines.append(f"classDef activity stroke:{options['stroke']}")
         if activity_stroke_width:
-            lines.append(f"classDef activity stroke-width:{options['activity']['stroke-width']}")
+            lines.append(
+                f"classDef activity stroke-width:{options['activity']['stroke-width']}"
+            )
         elif stroke_width:
             lines.append(f"classDef activity stroke-width:{options['stroke-width']}")
 
@@ -362,62 +462,96 @@ class ProvOntologyGraph():
         elif stroke_color:
             lines.append(f"classDef agent stroke:{options['stroke']}")
         if agent_stroke_width:
-            lines.append(f"classDef agent stroke-width:{options['agent']['stroke-width']}")
+            lines.append(
+                f"classDef agent stroke-width:{options['agent']['stroke-width']}"
+            )
         elif stroke_width:
             lines.append(f"classDef agent stroke-width:{options['stroke-width']}")
 
         for entity in self._entities:
             lines.append(
-                f"{entity.get_id()}{options['entity']['shape'].split(':')[0]}{entity.get_label() if entity.get_label() else entity.get_id()}{options['entity']['shape'].split(':')[1]}:::entity")
+                f'{entity.get_id()}{options["entity"]["shape"].split(":")[0]}<a style=color:inherit href={entity.get_id()}>{entity.get_label() if entity.get_label() else entity.get_id()}</a>{options["entity"]["shape"].split(":")[1]}:::entity'
+            )
             if "was_derived_by" in options["included-relations"]:
                 for item in entity._was_derived_from_entities:
                     if not options["revert-relations"]:
-                        lines.append(f"{entity.get_id()}-{entity_stroke_style if entity_stroke_style else stroke_style} was derived from {entity_stroke_style if entity_stroke_style else stroke_style}->{item.get_id()}")
+                        lines.append(
+                            f"{entity.get_id()}-{entity_stroke_style if entity_stroke_style else stroke_style} <a style=color:inherit href=https://www.w3.org/TR/prov-o/#wasDerivedFrom>was derived from</a> {entity_stroke_style if entity_stroke_style else stroke_style}->{item.get_id()}"
+                        )
                     else:
-                        lines.append(f"{item.get_id()}-{entity_stroke_style if entity_stroke_style else stroke_style} originated from {entity_stroke_style if entity_stroke_style else stroke_style}->{entity.get_id()}")
+                        lines.append(
+                            f"{item.get_id()}-{entity_stroke_style if entity_stroke_style else stroke_style} <a style=color:inherit href=https://www.w3.org/TR/prov-o/#wasDerivedFrom>originated from</a> {entity_stroke_style if entity_stroke_style else stroke_style}->{entity.get_id()}"
+                        )
             if "was_generated_by" in options["included-relations"]:
                 for item in entity._was_generated_by_activities:
                     if not options["revert-relations"]:
-                        lines.append(f"{entity.get_id()}-{activity_stroke_style if activity_stroke_style else stroke_style} was generated by {activity_stroke_style if activity_stroke_style else stroke_style}->{item.get_id()}")
+                        lines.append(
+                            f"{entity.get_id()}-{activity_stroke_style if activity_stroke_style else stroke_style} <a style=color:inherit href=https://www.w3.org/TR/prov-o/#wasGeneratedBy>was generated by</a> {activity_stroke_style if activity_stroke_style else stroke_style}->{item.get_id()}"
+                        )
                     else:
-                        lines.append(f"{item.get_id()}-{activity_stroke_style if activity_stroke_style else stroke_style} generated {activity_stroke_style if activity_stroke_style else stroke_style}->{entity.get_id()}")
+                        lines.append(
+                            f"{item.get_id()}-{activity_stroke_style if activity_stroke_style else stroke_style} <a style=color:inherit href=https://www.w3.org/TR/prov-o/#wasGeneratedBy>generated</a> {activity_stroke_style if activity_stroke_style else stroke_style}->{entity.get_id()}"
+                        )
             if "was_attributed_to" in options["included-relations"]:
                 for item in entity._was_attributed_to_agents:
                     if not options["revert-relations"]:
-                        lines.append(f"{entity.get_id()}-{agent_stroke_style if agent_stroke_style else stroke_style} was attributed to {agent_stroke_style if agent_stroke_style else stroke_style}->{item.get_id()}")
+                        lines.append(
+                            f"{entity.get_id()}-{agent_stroke_style if agent_stroke_style else stroke_style} <a style=color:inherit href=https://www.w3.org/TR/prov-o/#wasAttributedTo>was attributed to</a> {agent_stroke_style if agent_stroke_style else stroke_style}->{item.get_id()}"
+                        )
                     else:
-                        lines.append(f"{item.get_id()}-{agent_stroke_style if agent_stroke_style else stroke_style} produced {agent_stroke_style if agent_stroke_style else stroke_style}->{entity.get_id()}")
+                        lines.append(
+                            f"{item.get_id()}-{agent_stroke_style if agent_stroke_style else stroke_style} <a style=color:inherit href=https://www.w3.org/TR/prov-o/#wasAttributedTo>produced</a> {agent_stroke_style if agent_stroke_style else stroke_style}->{entity.get_id()}"
+                        )
 
         for activity in self._activities:
             lines.append(
-                f"{activity.get_id()}{options['activity']['shape'].split(':')[0]}{activity.get_label() if activity.get_label() else activity.get_id()}{options['activity']['shape'].split(':')[1]}:::activity")
+                f"{activity.get_id()}{options['activity']['shape'].split(':')[0]}<a style=color:inherit href={activity.get_id()}>{activity.get_label() if activity.get_label() else activity.get_id()}</a>{options['activity']['shape'].split(':')[1]}:::activity"
+            )
             if "was_informed_by" in options["included-relations"]:
                 for item in activity._was_informed_by_activities:
                     if not options["revert-relations"]:
-                        lines.append(f"{activity.get_id()}-{entity_stroke_style if entity_stroke_style else stroke_style} was informed by {entity_stroke_style if entity_stroke_style else stroke_style}->{item.get_id()}")
+                        lines.append(
+                            f"{activity.get_id()}-{entity_stroke_style if entity_stroke_style else stroke_style} <a style=color:inherit href=https://www.w3.org/TR/prov-o/#wasInformedBy>was informed by</a> {entity_stroke_style if entity_stroke_style else stroke_style}->{item.get_id()}"
+                        )
                     else:
-                        lines.append(f"{item.get_id()}-{entity_stroke_style if entity_stroke_style else stroke_style} informed {entity_stroke_style if entity_stroke_style else stroke_style}->{activity.get_id()}")
+                        lines.append(
+                            f"{item.get_id()}-{entity_stroke_style if entity_stroke_style else stroke_style} <a style=color:inherit href=https://www.w3.org/TR/prov-o/#wasInformedBy>informed</a> {entity_stroke_style if entity_stroke_style else stroke_style}->{activity.get_id()}"
+                        )
             if "used" in options["included-relations"]:
                 for item in activity._used_entities:
                     if not options["revert-relations"]:
-                        lines.append(f"{activity.get_id()}-{activity_stroke_style if activity_stroke_style else stroke_style} used {activity_stroke_style if activity_stroke_style else stroke_style}->{item.get_id()}")
+                        lines.append(
+                            f"{activity.get_id()}-{activity_stroke_style if activity_stroke_style else stroke_style} <a style=color:inherit href=https://www.w3.org/TR/prov-o/#used>used</a> {activity_stroke_style if activity_stroke_style else stroke_style}->{item.get_id()}"
+                        )
                     else:
-                        lines.append(f"{item.get_id()}-{activity_stroke_style if activity_stroke_style else stroke_style} was used by {activity_stroke_style if activity_stroke_style else stroke_style}->{activity.get_id()}")
+                        lines.append(
+                            f"{item.get_id()}-{activity_stroke_style if activity_stroke_style else stroke_style} <a style=color:inherit href=https://www.w3.org/TR/prov-o/#used>was used by</a> {activity_stroke_style if activity_stroke_style else stroke_style}->{activity.get_id()}"
+                        )
             if "was_associated_with" in options["included-relations"]:
                 for item in activity._was_associated_with_agents:
                     if not options["revert-relations"]:
-                        lines.append(f"{activity.get_id()}-{agent_stroke_style if agent_stroke_style else stroke_style} was associated with {agent_stroke_style if agent_stroke_style else stroke_style}->{item.get_id()}")
+                        lines.append(
+                            f"{activity.get_id()}-{agent_stroke_style if agent_stroke_style else stroke_style} <a style=color:inherit href=https://www.w3.org/TR/prov-o/#wasAssociatedWith>was associated with</a> {agent_stroke_style if agent_stroke_style else stroke_style}->{item.get_id()}"
+                        )
                     else:
-                        lines.append(f"{item.get_id()}-{agent_stroke_style if agent_stroke_style else stroke_style} initiated {agent_stroke_style if agent_stroke_style else stroke_style}->{activity.get_id()}")
+                        lines.append(
+                            f"{item.get_id()}-{agent_stroke_style if agent_stroke_style else stroke_style} <a style=color:inherit href=https://www.w3.org/TR/prov-o/#wasAssociatedWith>initiated</a> {agent_stroke_style if agent_stroke_style else stroke_style}->{activity.get_id()}"
+                        )
 
         for agent in self._agents:
-            lines.append(f"{agent.get_id()}{options['agent']['shape'].split(':')[0]}{agent.get_label() if agent.get_label() else agent.get_id()}{options['agent']['shape'].split(':')[1]}:::agent")
+            lines.append(
+                f"{agent.get_id()}{options['agent']['shape'].split(':')[0]}<a style=color:inherit href={agent.get_id()}>{agent.get_label() if agent.get_label() else agent.get_id()}</a>{options['agent']['shape'].split(':')[1]}:::agent"
+            )
             if "acted_on_behalf_of" in options["included-relations"]:
                 for item in agent._acted_on_behalf_of_agents:
                     if not options["revert-relations"]:
-                        lines.append(f"{agent.get_id()}-{agent_stroke_style if agent_stroke_style else stroke_style} acted on behalf of {agent_stroke_style if agent_stroke_style else stroke_style}->{item.get_id()}")
+                        lines.append(
+                            f"{agent.get_id()}-{agent_stroke_style if agent_stroke_style else stroke_style} <a style=color:inherit href=https://www.w3.org/TR/prov-o/#actedOnBehalfOf>acted on behalf of</a> {agent_stroke_style if agent_stroke_style else stroke_style}->{item.get_id()}"
+                        )
                     else:
-                        lines.append(f"{item.get_id()}-{agent_stroke_style if agent_stroke_style else stroke_style} instructed {agent_stroke_style if agent_stroke_style else stroke_style}->{agent.get_id()}")
+                        lines.append(
+                            f"{item.get_id()}-{agent_stroke_style if agent_stroke_style else stroke_style} <a style=color:inherit href=https://www.w3.org/TR/prov-o/#actedOnBehalfOf>instructed</a> {agent_stroke_style if agent_stroke_style else stroke_style}->{agent.get_id()}"
+                        )
 
         lines.append("```")
         lines = "\n".join(lines)
